﻿
using System;
using System.IO;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Json;
using System.Text;
#if (NETFRAMEWORK)
using System.Xaml;
#endif
using System.Xml.Serialization;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.IO.Extensions;
using OpcLabs.BaseLib.Utilities;

namespace OpcLabs.UnitTestLib
{
    [System.Runtime.InteropServices.GuidAttribute("26DB4974-5AB7-4B90-8D79-378EC5ECA94D")]
    static public class Tests
    {
        static public void BinarySerializationTest<T>([NotNull] T target)
        {
            BinarySerializationTest(target, (arg1, arg2) => Assert.IsTrue(ObjectUtilities.ValueEquals(arg1, arg2), "The deserialized object is not value equal to the original."));
        }

        static public void BinarySerializationTest<T>([NotNull] T target, [NotNull] Action<T, T> assertion)
        {
            using (var memoryStream = new MemoryStream())
            {
                memoryStream.BinarySerialize(target);

                AssertUtilities.WriteLine($"Stream length: {memoryStream.Length}");

                var actual = memoryStream.BinaryDeserialize<T>();
                assertion(target, actual);
            }
        }

        static public void DataContractJsonSerializationTest<T>([NotNull] T target)
        {
            DataContractJsonSerializationTest(target, (arg1, arg2) => Assert.IsTrue(ObjectUtilities.ValueEquals(arg1, arg2), "The deserialized object is not value equal to the original."));
        }

        static public void DataContractJsonSerializationTest<T>([NotNull] T target, [NotNull] Action<T, T> assertion)
        {
            using (var memoryStream = new MemoryStream())
            {
                var serializer = new DataContractJsonSerializer(target.GetType()/*, knownTypes*/);
                serializer.WriteObject(memoryStream, target);

                memoryStream.Position = 0;
                using (var streamReader = new StreamReader(memoryStream))
                {
                    AssertUtilities.WriteLine(streamReader.ReadToEnd());

                    memoryStream.Position = 0;
                    var actual = (T)serializer.ReadObject(memoryStream);
                    assertion(target, actual);
                }
            }
        }

        static public void DataContractSerializationTest<T>([NotNull] T target)
        {
            DataContractSerializationTest(target, (arg1, arg2) => Assert.IsTrue(ObjectUtilities.ValueEquals(arg1, arg2), "The deserialized object is not value equal to the original."));
        }

        static public void DataContractSerializationTest<T>([NotNull] T target, [NotNull] Action<T, T> assertion)
        {
            using (var memoryStream = new MemoryStream())
            {
                var serializer = new DataContractSerializer(target.GetType()/*, knownTypes*/);
                serializer.WriteObject(memoryStream, target);

                memoryStream.Position = 0;
                using (var streamReader = new StreamReader(memoryStream))
                {
                    AssertUtilities.WriteLine(streamReader.ReadToEnd());

                    memoryStream.Position = 0;
                    var actual = (T) serializer.ReadObject(memoryStream);
                    assertion(target, actual);
                }
            }
        }

#if (NETFRAMEWORK)
        static public void XamlSerializationTest<T>([NotNull] T target)
        {
            XamlSerializationTest(target, (arg1, arg2) => Assert.IsTrue(ObjectUtilities.ValueEquals(arg1, arg2), "The deserialized object is not value equal to the original."));
        }
#endif

#if (NETFRAMEWORK)
        static public void XamlSerializationTest<T>([NotNull] T target, [NotNull] Action<T, T> assertion)
        {
            var stringBuilder = new StringBuilder();
            using (TextWriter textWriter = new StringWriter(stringBuilder))
                XamlServices.Save(textWriter, target);

            string serializedText = stringBuilder.ToString();
            AssertUtilities.WriteLine(serializedText);

            using (TextReader textReader = new StringReader(serializedText))
            {
                var actual = (T)(XamlServices.Load(textReader));
                assertion(target, actual);
            }
        }
#endif

        static public void XmlSerializationTest<T>([NotNull] T target)
        {
            XmlSerializationTest(target, (arg1, arg2) => Assert.IsTrue(ObjectUtilities.ValueEquals(arg1, arg2), "The deserialized object is not value equal to the original."));
        }

        static public void XmlSerializationTest<T>([NotNull] T target, [NotNull] Action<T, T> assertion)
        {
            var xmlSerializer = new XmlSerializer(target.GetType());
            var stringBuilder = new StringBuilder();
            using (TextWriter textWriter = new StringWriter(stringBuilder))
                xmlSerializer.Serialize(textWriter, target);

            string serializedText = stringBuilder.ToString();
            AssertUtilities.WriteLine(serializedText);

            using (TextReader textReader = new StringReader(serializedText))
            {
                var actual = (T) (xmlSerializer.Deserialize(textReader));
                assertion(target, actual);
            }
        }
    }
}
