﻿using System;
using System.Collections.Generic;
using System.Linq;
using JetBrains.Annotations;
using OpcLabs.BaseLib.Browsing;
using OpcLabs.BaseLib.Diagnostics;

namespace OpcLabs.BaseLibFormsTestApp
{
    class BrowseableTestSample : IBrowseable
    {
        public void BrowseBranches(object parentNode, object nodeFilter, out IEnumerable<object> branches, out IEnumerable<Exception> exceptions)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            Check.Ensures(Check.ValueAtReturn(out branches) != null);
            Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

            if (parentNode == null) throw new ArgumentNullException(nameof(parentNode));

            if (parentNode.GetType().Name == "Root")
            {
                System.Threading.Thread.Sleep(1000+1000);
                branches = new Network[] { "First", "Second", "Three" };
                exceptions = Enumerable.Empty<Exception>();
                return;
            }

            if (parentNode.GetType().Name == "Network")
            {
                System.Threading.Thread.Sleep(2000);
//                branches = Filter(new Folder[] { "1", "2", "3", "02" }, nodeFilter);
                branches = Filter(new Folder[] { "1", "2", "3", "02", "111a", "1b", "1a", "1c", "1e", "1F", "10", "1D", "1s", "1y", "0o", "0l", "0i", "7", "6" }, nodeFilter);
                exceptions = Enumerable.Empty<Exception>();
                return;
            }

            if (parentNode.GetType().Name == "Folder")
            {
                System.Threading.Thread.Sleep(0+3000);
                // ReSharper disable LocalizableElement
                if (parentNode.ToString() == "3")
                {
                    branches = Enumerable.Empty<Object>();
                    branches = new[] { "EEE" }; // for testing bug after expand node in tree and return exception and non empty branches
                    exceptions = new[] { new Exception("Timeout.\nxyz\nščřž") };
                    return;
                }
                // ReSharper restore LocalizableElement
                branches = new[] { "A", "B" };
                exceptions = Enumerable.Empty<Exception>();
                return;
            }
            branches = Enumerable.Empty<Object>();
            exceptions = Enumerable.Empty<Exception>();
        }

        public void BrowseLeaves(object parentNode, object nodeFilter, out IEnumerable<object> leaves, out IEnumerable<Exception> exceptions)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            Check.Ensures(Check.ValueAtReturn(out leaves) != null);
            Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

            if (parentNode == null) throw new ArgumentNullException(nameof(parentNode));

            if (parentNode.GetType().Name == "Root")
            {
                leaves = Enumerable.Empty<Object>();
                exceptions = Enumerable.Empty<Exception>();
                return;
            }
            if ((parentNode is Network s) && (s.ToString() != "Second"))
            {
                System.Threading.Thread.Sleep(3000);
                WorkGroup wg = s + " Group";
                var f02 = new Folder("02"); // leaf and branch
                leaves = new object[] { s + "Name", wg, f02 };
                exceptions = Enumerable.Empty<Exception>();
                return;
            }
            if (parentNode is Folder f)
            {
                System.Threading.Thread.Sleep(500);
                // ReSharper disable LocalizableElement
                if (f.Name == "3")
                {
                    leaves = new object[] { "L1" };

                    var r = new System.Windows.Forms.RichTextBox();
                    r.SelectionStart = r.TextLength;
                    r.SelectionColor = System.Drawing.Color.Red;
                    r.SelectedText = "Cannot read \t";
                    r.SelectionStart = r.TextLength;
                    r.SelectionColor = System.Drawing.Color.Blue;
                    r.SelectionFont = new System.Drawing.Font("Arial", 14, System.Drawing.FontStyle.Bold);
                    r.SelectionBackColor = System.Drawing.Color.Magenta;
                    r.SelectionBullet = true;
                    r.SelectedText = "Timeout" + "\n";
                    r.AppendText("ěščřžýáíé www.opclabs.com");

                    exceptions = new Exception[] { new BrowseException(r.Rtf, new BrowseException("Inner exception. See: www.opclabs.com")) };
                    return;
                }
                // ReSharper restore LocalizableElement
                leaves = new object[] { "L1", "L2", "L0", "L012", "L01a20", "L1a30", "L01a30" };
                exceptions = Enumerable.Empty<Exception>();
                return;
            }
            leaves = Enumerable.Empty<Object>();
            exceptions = Enumerable.Empty<Exception>();
        }

        public void CreateNode(object parentNode, Type type, string name, out object node, out Exception exception)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            Always.RequiresNotNull(type, nameof(type));
            Always.RequiresNotNull(name, nameof(name));
            Check.Ensures(Check.ValueAtReturn(out node) != null || Check.ValueAtReturn(out exception) != null);

            try
            {
                node = Activator.CreateInstance(type, new object[] { name });
                exception = null;
            }
            catch (Exception e)
            {
                node = null;
                exception = e;
            }
        }

        public IEnumerable<Type> GetAddableBranchTypes(object parentNode, bool onlyUserAddable)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            Check.EnsuresResultNotNull<IEnumerable<Type>>();

            var root = parentNode as Root;
            var network = parentNode as Network;
            var folder = parentNode as Folder;
            if (root != null)
            {
                network = new Network("");
                return new[] { network.GetType() };
            }
            if (network != null)
            {
                var newFolder = new Folder("");
                return new[] { newFolder.GetType() };
            }
            if (folder != null)
            {
                var workgroup = new WorkGroup("");
                return new[] { workgroup.GetType() };
            }
            return Enumerable.Empty<Type>();

        }

        public IEnumerable<Type> GetAddableLeafTypes(Object parentNode, bool onlyUserAddable)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            Check.EnsuresResultNotNull<IEnumerable<Type>>();

            var network = parentNode as Network;
            var folder = parentNode as Folder;
            //if (folder != null)
            //{
            //    const string s = "";
            //    return new[] { s.GetType() };
           // }
            if (network != null)
            {
                var wg = new WorkGroup("xxx");
                return new[] { wg.GetType() };
            }
            return Enumerable.Empty<Type>();
        }

        public void HasChildren(object parentNode, out bool parentHasBranches, out bool parentHasLeaves, out IEnumerable<Exception> exceptions)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            Check.Ensures(Check.ValueAtReturn(out exceptions) != null);

            if (parentNode == null) throw new ArgumentNullException(nameof(parentNode));
            parentHasBranches = parentNode.ToString() != "1";
            parentHasLeaves = parentNode.ToString() != "1";
            exceptions = Enumerable.Empty<Exception>();
        }

        public void InvalidateCache(object parentNode)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            //throw new NotImplementedException();
        }

        public bool IsBranch(object parentNode, object node)
        {
            Always.RequiresNotNull(node, nameof(node));
            return true;
            //throw new NotImplementedException();
        }

        public bool IsLeaf(object parentNode, object node)
        {
            Always.RequiresNotNull(node, nameof(node));
            return true;
            //throw new NotImplementedException();
        }

        public bool IsMyNode(object parentNode, object node)
        {
            Always.RequiresNotNull(node, nameof(node));
            return true;
        }

        public bool IsSelectable(object parentNode, object node)
        {
            Always.RequiresNotNull(node, nameof(node));
            return true;
            //throw new NotImplementedException();
        }

        [NotNull]
        private IEnumerable<Object> Filter([NotNull] IEnumerable<Object> objects, [CanBeNull] Object nodeFilter)
        {
            if (nodeFilter == null) return objects;
            IEnumerable<Object> result = objects.Where(obj => Equals(obj, nodeFilter));
            return result;
        }

        #region IBrowseable Members


        public bool MatchNode(object parentNode, object nodeA, object nodeB)
        {
            Always.RequiresNotNull(nodeA, nameof(nodeA));
            Always.RequiresNotNull(nodeB, nameof(nodeB));

            return Equals(nodeB, nodeA);
        }

        public void NodeAdded(object parentNode, object node)
        {
            Always.RequiresNotNull(parentNode, nameof(parentNode));
            Always.RequiresNotNull(node, nameof(node));
            // do nothing
        }

        #endregion
    }

    class Root
    {
        public Root(string name)
        {
            Name = name;
        }

        public string Name { get; set; }

        /// <inheritdoc cref="object.ToString()"/>
        public override string ToString()
        {
            return Name ?? "";
        }

        /// <inheritdoc cref="Object.GetHashCode"/>
        public override int GetHashCode()
        {
            return Name == null ? 0 : Name.GetHashCode();
        }

        /// <inheritdoc cref="Object.Equals(object)"/>
        public override bool Equals(object obj)
        {
            if (obj is Root root)
            {
                if (String.Equals(root.Name, "*")) return true;
                return String.Equals(root.Name, Name);
            }
            return base.Equals(obj);
        }

        static public implicit operator Root(string value)
        {
            return new Root(value);
        }
    }

    class Folder
    {
        public Folder(string name)
        {
            Name = name;
        }

        public string Name { get; set; }

        /// <inheritdoc cref="object.ToString()"/>
        public override string ToString()
        {
            if (Name == null) return "";
            return Name;
        }

        /// <inheritdoc cref="Object.GetHashCode"/>
        public override int GetHashCode()
        {
            return Name == null ? 0 : Name.GetHashCode();
        }

        static public implicit operator Folder(string value)
        {
            return new Folder(value);
        }

        /// <inheritdoc cref="Object.Equals(object)"/>
        public override bool Equals(object obj)
        {
            if (obj is Folder folder)
            {
                if (String.Equals(folder.Name, "*")) return true;
                return String.Equals(folder.Name, Name);
            }
            return base.Equals(obj);
        }
    }

    class FolderFilter : Folder, ICloneable
    {
        public FolderFilter(string name)
            : base(name)
        {
        }

   
        #region ICloneable Members

        public object Clone()
        {
            return new FolderFilter(Name); 
        }

        #endregion
}

    class Network
    {
        public Network(string name)
        {
            Name = name;
        }

        public string Name { get; set; }

        /// <inheritdoc cref="object.ToString()"/>
        public override string ToString()
        {
            if (Name == null) return "";
            return Name;
        }

        /// <inheritdoc cref="Object.GetHashCode"/>
        public override int GetHashCode()
        {
            return Name == null ? 0 : Name.GetHashCode();
        }

        /// <inheritdoc cref="Object.Equals(object)"/>
        public override bool Equals(object obj)
        {
            if (obj is Network network)
            {
                if (String.Equals(network.Name, "*")) return true;
                return String.Equals(network.Name, Name);
            }
            return base.Equals(obj);
        }

        static public implicit operator Network(string value)
        {
            return new Network(value);
        }
    }

    class WorkGroup
    {
        public WorkGroup(string name)
        {
            Name = name;
        }

        public string Name { get; set; }

        /// <inheritdoc cref="object.ToString()"/>
        public override string ToString()
        {
            return Name ?? "";
        }

        /// <inheritdoc cref="Object.Equals(object)"/>
        public override bool Equals(object obj)
        {
            if (obj is WorkGroup workGroup)
            {
                if (String.Equals(workGroup.Name, "*")) return true;
                return String.Equals(workGroup.Name, Name);
            }
            return base.Equals(obj);
        }

        /// <inheritdoc cref="Object.GetHashCode"/>
        public override int GetHashCode()

        {
            return Name == null ? 0 : Name.GetHashCode();
        }

        static public implicit operator WorkGroup(string value)
        {
            return new WorkGroup(value);
        }

    }

    /*
    class BrowseableItem : IBrowseable
    {
    }*/
}
