﻿
using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Navigation;
using OpcLabs.BaseLib.Runtime.InteropServices;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.AlarmsAndEvents;
using OpcLabs.EasyOpc.AlarmsAndEvents.AddressSpace;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.AlarmsAndEvents
{
    partial class EasyAEClientTest
    {
        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_OpcAeScheme()
        {
            ServerDescriptor serverDescriptor = AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId;
            serverDescriptor.Scheme = ServerDescriptor.OpcAEScheme;
            BrowseNodesTest_Basic(serverDescriptor, AENodeDescriptor.Root, new AEBrowseParameters(AEBrowseFilter.All), 1);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Unknown1Scheme()
        {
            ServerDescriptor serverDescriptor = AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId;
            serverDescriptor.Scheme = "unknown1";

            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyAEClient();

            try
            {
                client.BrowseNodes(serverDescriptor, AENodeDescriptor.Root, new AEBrowseParameters(AEBrowseFilter.All));
            }
            catch (OpcException opcException)
            {
                Exception innerException = opcException.InnerException;
                Assert.IsNotNull(innerException);
                Assert.IsInstanceOfType(innerException, typeof(NotSupportedException));
                return;
            }
            Assert.Fail();
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Root_Both_Areas()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, AENodeDescriptor.Root, new AEBrowseParameters(AEBrowseFilter.Areas), 1);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Root_BrowsePath_Areas()
        {
            try
            {
                BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, new AENodeDescriptor { BrowsePath = BrowsePath.Root }, new AEBrowseParameters(AEBrowseFilter.Areas), 1);
            }
            catch (OpcException opcException)
            {
                AssertUtilities.WriteLine($"opcException: {opcException}");
                AssertUtilities.WriteLine($"opcException.InnerException: {opcException.InnerException}");
                Assert.IsInstanceOfType(opcException.InnerException, typeof(NotImplementedException));
                return;
            }
            Assert.Fail();
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Root_QualifiedName_Areas()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, "", new AEBrowseParameters(AEBrowseFilter.Areas), 1);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Root_Both_All()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, AENodeDescriptor.Root, new AEBrowseParameters(AEBrowseFilter.All), 1);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Root_Both_Sources()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, AENodeDescriptor.Root, new AEBrowseParameters(AEBrowseFilter.Sources), 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Simulation_Areas()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, "Simulation", new AEBrowseParameters(AEBrowseFilter.Areas), 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Simulation_All()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, "Simulation", new AEBrowseParameters(AEBrowseFilter.All), 34);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_Simulation_Sources()
        {
            BrowseNodesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, "Simulation", new AEBrowseParameters(AEBrowseFilter.Sources), 34);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void BrowseNodesTest_Basic_UnknownNodeId()
        {
            BrowseNodesTest_UnknownNodeId(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void QueryEventCategoriesTest_Basic_All()
        {
            QueryEventCategoriesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, 
                AEEventTypes.All, 14);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void QueryEventCategoriesTest_Basic_Condition()
        {
            QueryEventCategoriesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                AEEventTypes.Condition, 8);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void QueryEventCategoriesTest_Basic_Condition_Simple()
        {
            QueryEventCategoriesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                AEEventTypes.Condition | AEEventTypes.Simple, 11);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void QueryEventCategoriesTest_Basic_None()
        {
            QueryEventCategoriesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                AEEventTypes.None, 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void QueryEventCategoriesTest_Basic_Simple()
        {
            QueryEventCategoriesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                AEEventTypes.Simple, 3);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void QueryEventCategoriesTest_Basic_Tracking()
        {
            QueryEventCategoriesTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                AEEventTypes.Tracking, 3);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void QuerySourceConditionsTest_Basic()
        {
            QuerySourceConditionsTest_Basic(AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                "Simulation.ConditionState1", 1);
        }

        #endregion



        #region Protected Methods

        protected void BrowseNodesTest_UnknownNodeId([NotNull] ServerDescriptor serverDescriptor)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyAEClient();

            AENodeElementCollection actualResultCollection;
            try
            {
                TestSection.Act();
                try
                {
                    actualResultCollection = client.BrowseAreas(serverDescriptor, "unknownNodeId");
                }
                finally
                {
                    TestSection.Assert();
                }
            }
            catch (OpcException opcException)
            {
                Exception innerException = opcException.InnerException;
                Assert.IsNotNull(innerException);

                Exception baseException = innerException.GetBaseException();
                Assert.IsNotNull(baseException);
                AssertUtilities.WriteLine($"baseException: {baseException}");
                Assert.IsInstanceOfType(baseException, typeof(ExtendedCOMException));
                var extendedComException = (ExtendedCOMException)baseException;
                AssertUtilities.WriteLine($"extendedComException.QualifiedName: {extendedComException.QualifiedName}");
                AssertUtilities.WriteLine($"extendedComException.GetErrorId(): {extendedComException.GetErrorId()}");
                // ReSharper disable once CommentTypo
                Assert.AreEqual(unchecked((int)0xC0040203), extendedComException.ErrorCode);   // OPC_E_INVALIDBRANCHNAME
                Assert.AreEqual("http://opcfoundation.org/AlarmAndEvents/", extendedComException.QualifiedName.Namespace);
                // ReSharper disable once StringLiteralTypo
                Assert.AreEqual("E_INVALIDBRANCHNAME", extendedComException.QualifiedName.Name);

                return;
            }
            AssertUtilities.WriteLine(actualResultCollection.Count);
            Assert.Fail();
        }

        #endregion



        #region Private Methods

        private void BrowseNodesTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] AENodeDescriptor parentNodeDescriptor,
            [NotNull] AEBrowseParameters browseParameters,
            int expectedCount)  // -1 for "unknown"
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine($"parentNodeDescriptor: {parentNodeDescriptor}");
            TestEnvironment.PrecheckServer(serverDescriptor);
            
            var client = new EasyAEClient();

            TestSection.Act();
            AENodeElementCollection nodeElements = client.BrowseNodes(
                serverDescriptor,
                parentNodeDescriptor,
                browseParameters);

            TestSection.Assert();
            Assert.IsNotNull(nodeElements);
            AssertUtilities.WriteLine(nodeElements.Count);
            CollectionAssert.AllItemsAreNotNull(nodeElements);
            foreach (AENodeElement nodeElement in nodeElements)
                AssertUtilities.WriteLine($"nodeElement: {nodeElement}");
            if (expectedCount >= 0)
                Assert.AreEqual(expectedCount, nodeElements.Count);
        }

        private void QueryEventCategoriesTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] AEEventType eventTypes,
            int expectedCount)  // -1 for "unknown"
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyAEClient();

            TestSection.Act();
            AECategoryElementCollection categoryElements = client.QueryEventCategories(serverDescriptor, eventTypes);

            TestSection.Assert();
            Assert.IsNotNull(categoryElements);
            AssertUtilities.WriteLine(categoryElements.Count);
            CollectionAssert.AllItemsAreNotNull(categoryElements);
            foreach (AECategoryElement categoryElement in categoryElements)
                AssertUtilities.WriteLine($"categoryElement: {categoryElement}");
            if (expectedCount >= 0)
                Assert.AreEqual(expectedCount, categoryElements.Count);
        }

        private void QuerySourceConditionsTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] AENodeDescriptor sourceDescriptor,
            int expectedCount)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyAEClient();

            TestSection.Act();
            AEConditionElementCollection conditionElements = client.QuerySourceConditions(serverDescriptor, sourceDescriptor);

            TestSection.Assert();
            Assert.IsNotNull(conditionElements);
            AssertUtilities.WriteLine(conditionElements.Count);
            CollectionAssert.AllItemsAreNotNull(conditionElements);
            foreach (AEConditionElement conditionElement in conditionElements)
            {
                AssertUtilities.WriteLine($"conditionElement: {conditionElement}");
                Assert.IsNotNull(conditionElement);
                Assert.IsNotNull(conditionElement.Name);
                Assert.IsNotNull(conditionElement.SubconditionNames);
            }
            Assert.AreEqual(expectedCount, conditionElements.Count);
        }

        #endregion
    }
}
