﻿
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.AlarmsAndEvents;
using OpcLabs.EasyOpc.AlarmsAndEvents.AddressSpace;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.AlarmsAndEvents
{
    [TestClass]
    public partial class EasyAEClientTest
    {
        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestMethod]
        public void GetConditionStateTest_Basic()
        {
            GetConditionStateTest_Basic(
                AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                "Simulation.ConditionState1",
                "Simulated",
                new long[0]);
        }

        #endregion



        #region Protected Properties

        [NotNull]
        protected AELiveTestEnvironment TestEnvironment => AELiveTestEnvironment.Current;

        #endregion



        #region Private Methods

        private void GetConditionStateTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] AENodeDescriptor sourceDescriptor,
            [NotNull] string conditionName,
            [NotNull] long[] attributes)
        {
            TestSection.Arrange();
            TestEnvironment.PrecheckServer(serverDescriptor);
            var client = new EasyAEClient();

            TestSection.Act();
            AEConditionState conditionState = client.GetConditionState(serverDescriptor, sourceDescriptor, conditionName, attributes);

            TestSection.Assert();
            Assert.IsNotNull(conditionState);
            AssertUtilities.WriteLine(conditionState);

            AEAttributeValueDictionary attributeValues = conditionState.AttributeValues;
            Assert.IsNotNull(attributeValues);
            Assert.AreEqual(attributes.Length, attributeValues.Count);
            Assert.AreEqual(new DAQuality(DAQualities.GoodNonspecific), conditionState.Quality);
            AESubconditionElementCollection subconditionElements = conditionState.SubconditionElements;
            Assert.IsNotNull(subconditionElements);
            Assert.AreEqual(1, subconditionElements.Count);
            AESubconditionElement subconditionElement = subconditionElements[0];
            Assert.IsNotNull(subconditionElement);
            Assert.AreEqual("(no definition)", subconditionElement.Definition);
            Assert.AreEqual("(no description)", subconditionElement.Description);
            Assert.AreEqual("Simulated", subconditionElement.Name);
            Assert.AreEqual(1, subconditionElement.Severity);

            // IMPROVE: test more
        }

        #endregion
    }
}
