﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Runtime.InteropServices;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.ComInterop;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.AddressSpace;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    // ReSharper disable once InconsistentNaming
    public partial class IEasyDAClientTest_Com
    {

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseBranchesTest()
        {
            var target = CreateClient();

            DANodeElementCollection actualResultCollection;
            try
            {
                actualResultCollection = target.BrowseBranches("", OpcLabs_KitServer.ProgId);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.IsNotNull(actualResultCollection);

            foreach (DANodeElement nodeElement in actualResultCollection)
            {
                Assert.IsNotNull(nodeElement);
                AssertUtilities.WriteLine(nodeElement);
            }

            //TODO:

        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest()
        {
            var target = CreateClient();

            DAPropertyElementCollection actualResultCollection;
            try
            {
                actualResultCollection = target.BrowseProperties("", OpcLabs_KitServer.ProgId,
                    OpcLabs_KitServer.ItemIds.
                        Constant_I4);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.IsNotNull(actualResultCollection);
            Assert.IsTrue(actualResultCollection.Count >= 12);

            //TODO:

        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest_ServerDescriptor()
        {
            var target = CreateClient();

            DAPropertyElementCollection actualResultCollection;
            try
            {
                actualResultCollection = target.BrowseProperties(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId), OpcLabs_KitServer.ItemIds.Constant_I4);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.IsNotNull(actualResultCollection);
            Assert.IsTrue(actualResultCollection.Count >= 12);

            //TODO:

        }

        [ExpectedException(typeof(OpcException), AllowDerivedTypes = true)]
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowsePropertiesTest_UnknownItemId_KitServer()
        {
            var target = CreateClient();

            DAPropertyElementCollection actualResultCollection = target.BrowseProperties(
                OpcLabs_KitServer.ProgId,
                "unknownItemId");
            AssertUtilities.WriteLine(actualResultCollection.Count);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetMultiplePropertyValuesTest()
        {
            var target = CreateClient();

            var propertyArgumentsArray = new[]
                {
                    new DAPropertyArguments(OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_I4, DAPropertyIds.DataType), 
                    new DAPropertyArguments(OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_I4, DAPropertyIds.AccessRights), 
                    new DAPropertyArguments(OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4, DAPropertyIds.DataType), 
                    new DAPropertyArguments(OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4, DAPropertyIds.AccessRights), 
                    new DAPropertyArguments(OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4, DAPropertyIds.Description)
                };

            ValueResult[] resultArray = target.GetMultiplePropertyValues(propertyArgumentsArray);
            Assert.IsNotNull(resultArray);
            Assert.AreEqual(5, resultArray.Length);
            CollectionAssert.AllItemsAreNotNull(resultArray);

            if (!target.SupportsTechnology(OpcTechnologies.Com))
            {
                foreach (ValueResult valueResult in resultArray)
                    // ReSharper disable once PossibleNullReferenceException
                    Assert.IsInstanceOfType(valueResult.Exception, typeof(NotSupportedException));
                return;
            }

            ValueResult result0 = resultArray[0];
            Assert.IsNotNull(result0);
            AssertUtilities.WriteLine(result0.Exception);
            Assert.IsNull(result0.Exception);
            Assert.AreEqual((short)VarTypes.I4, result0.Value);

            ValueResult result1 = resultArray[1];
            Assert.IsNotNull(result1);
            Assert.IsNull(result1.Exception);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreEqual((int)DAAccessRights.Readable, (int)result1.Value);
            // ReSharper restore PossibleNullReferenceException

            ValueResult result2 = resultArray[2];
            Assert.IsNotNull(result2);
            Assert.IsNull(result2.Exception);
            Assert.AreEqual((short)VarTypes.R4, result2.Value);

            ValueResult result3 = resultArray[3];
            Assert.IsNotNull(result3);
            Assert.IsNull(result3.Exception);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreEqual((int)DAAccessRights.Readable, (int)result3.Value);
            // ReSharper restore PossibleNullReferenceException

            ValueResult result4 = resultArray[4];
            Assert.IsNotNull(result4);
            Assert.IsNotNull(result4.Exception);
            AssertUtilities.WriteLine(result4.Exception);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void DisposeTest_ItemChangedAbsence()
        {
            UnsubscribeAllItemsTest_ItemChangedAbsence(
                CreateClient,
                // ReSharper disable once PossibleNullReferenceException
                client => ((IDisposable)client).Dispose());
        }

        [TestCategory("Live")]
        [TestMethod]
        public void GetPropertyValueTest()
        {
            var target = CreateClient();

            object actualValue;
            try
            {
                actualValue = target.GetPropertyValue("", OpcLabs_KitServer.ProgId,
                    OpcLabs_KitServer.ItemIds.Constant_I4, DAPropertyIds.DataType);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.AreEqual((short)VarTypes.I4, actualValue);

            //TODO:

        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetPropertyValueTest_ServerDescriptor()
        {
            var target = CreateClient();

            object actualValue;
            try
            {
                actualValue = target.GetPropertyValue(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    OpcLabs_KitServer.ItemIds.Constant_I4,
                    DAPropertyIds.DataType);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            Assert.AreEqual((short)VarTypes.I4, actualValue);

            //TODO:

        }

        ///<summary>
        ///A test for ReadItem(System::String ^, System::String ^, System::String ^)
        ///</summary>
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        // ReSharper disable InconsistentNaming
        public void ReadItemTest_Constant_BOOL()
        // ReSharper restore InconsistentNaming
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_BOOL, false);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        // ReSharper disable InconsistentNaming
        public void ReadItemTest_Constant_BSTR()
        // ReSharper restore InconsistentNaming
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_BSTR, "");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_Constant_I2()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_I2, (short)0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_Constant_I4()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_I4, 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void ReadItemTest_Constant_I4_AfterDelay()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_I4, 0);
            // The intent of the delay is to force an OPC read (and not just using value from subscription)
            AssertUtilities.WaitingPhase(3 * 1000);
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_I4, 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_Constant_R4()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4, 0.0F);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_Constant_R8()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R8, 0.0D);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_DataType_R8()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4,
                         Convert.ToInt16(System.Runtime.InteropServices.VarEnum.VT_R8), 0.0D);
        }

        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_DataType_Invalid1()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4,
                         DALiveTestDefinitions.DataTypes.Invalid1, 0.0F);
        }

        [ExpectedException(typeof(ArgumentOutOfRangeException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_DataType_Invalid2()
        {
            ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4,
                DALiveTestDefinitions.DataTypes.Invalid2, 0.0F);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_MachineName_Null()
        {
            var target = CreateClient();
            // ReSharper disable AssignNullToNotNullAttribute
            target.ReadItem(null, OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Constant_R4);
            // ReSharper restore AssignNullToNotNullAttribute
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_ServerClass_Null()
        {
            var target = CreateClient();
            // ReSharper disable AssignNullToNotNullAttribute
            target.ReadItem("", null, OpcLabs_KitServer.ItemIds.Constant_R4);
            // ReSharper restore AssignNullToNotNullAttribute
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_ItemId_Null()
        {
            var target = CreateClient();
            // ReSharper disable AssignNullToNotNullAttribute
            target.ReadItem("", OpcLabs_KitServer.ProgId, null);
            // ReSharper restore AssignNullToNotNullAttribute
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_UnknownItem()
        {
            try
            {
                if (!ReadItemTest("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.UnknownItem, 0.0F))
                    return;
            }
            catch (OpcException ex)
            {
                Exception innerException = ex.InnerException;
                Assert.IsNotNull(innerException);
                var comException = innerException as COMException;
                Assert.IsNotNull(comException);
                Assert.AreEqual(-1073479673, comException.ErrorCode);
                // The item is no longer available in the server address space
                return;
            }
            Assert.Fail();
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_Descriptors()
        {
            ReadItemTest(new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                         new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Constant_R4), 0.0F);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_ServerDescriptor_Null()
        {
            var target = CreateClient();
            // ReSharper disable AssignNullToNotNullAttribute
            target.ReadItem(null, new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Constant_R4));
            // ReSharper restore AssignNullToNotNullAttribute
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTest_ItemDescriptor_Null()
        {
            var target = CreateClient();
            // ReSharper disable AssignNullToNotNullAttribute
            target.ReadItem(new ServerDescriptor("", OpcLabs_KitServer.ProgId), null);
            // ReSharper restore AssignNullToNotNullAttribute
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueTest_Constant_R4()
        {
            var target = CreateClient();
            object actualValue;
            try
            {
                actualValue = target.ReadItemValue("", OpcLabs_KitServer.ProgId,
                    OpcLabs_KitServer.ItemIds.Constant_R4);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualValue);
            Assert.AreEqual(0.0F, actualValue);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueTest_Constant_R8()
        {
            var target = CreateClient();
            object actualValue;
            try
            {
                actualValue = target.ReadItemValue("", OpcLabs_KitServer.ProgId,
                    OpcLabs_KitServer.ItemIds.Constant_R8);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualValue);
            Assert.AreEqual(0.0D, actualValue);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueTest_DataType_R8()
        {
            var target = CreateClient();
            object actualValue;
            try
            {
                actualValue = target.ReadItemValue("", OpcLabs_KitServer.ProgId,
                    OpcLabs_KitServer.ItemIds.Constant_R4,
                    Convert.ToInt16(VarEnum.VT_R8));
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualValue);
            Assert.AreEqual(0.0D, actualValue);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueTest_ServerDescriptor_Null()
        {
            var target = CreateClient();
            // ReSharper disable AssignNullToNotNullAttribute
            object actualValue;
            try
            {
                actualValue = target.ReadItemValue(null,
                    new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Constant_R4));
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            // ReSharper restore AssignNullToNotNullAttribute
            AssertUtilities.WriteLine(actualValue);
        }

        [ExpectedException(typeof(ArgumentNullException))]
        [TestCategory("ArgumentChecking")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueTest_ItemDescriptor_Null()
        {
            var target = CreateClient();
            // ReSharper disable AssignNullToNotNullAttribute
            object actualValue = target.ReadItemValue(new ServerDescriptor("", OpcLabs_KitServer.ProgId), null);
            // ReSharper restore AssignNullToNotNullAttribute
            AssertUtilities.WriteLine(actualValue);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemValueTest_UnknownItem()
        {
            var target = CreateClient();
            object actualValue;
            try
            {
                try
                {
                    actualValue = target.ReadItemValue("", OpcLabs_KitServer.ProgId,
                        OpcLabs_KitServer.ItemIds.UnknownItem);
                }
                catch (OpcException opcException)
                {
                    if (!target.SupportsTechnology(OpcTechnologies.Com))
                    {
                        Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                        return;
                    }
                    throw;
                }
            }
            catch (OpcException ex)
            {
                Exception innerException = ex.InnerException;
                Assert.IsNotNull(innerException);
                var comException = innerException as COMException;
                Assert.IsNotNull(comException);
                Assert.AreEqual(-1073479673, comException.ErrorCode);
                // The item is no longer available in the server address space
                return;
            }
            AssertUtilities.WriteLine(actualValue);
            Assert.Fail();
        }

        // Allowing null arguments so that we can test Requires contract violations.
        static public void SubscribeItemTest(
            [NotNull] IEasyDAClient client,
            [CanBeNull] ServerDescriptor serverDescriptor, 
            [CanBeNull] DAItemDescriptor itemDescriptor,
            bool unsubscribeAllItems)
        {
            int successCount = 0;
            int failureCount = 0;
            void ItemChanged(object sender, EasyDAItemChangedEventArgs e)
            {
                AssertUtilities.WriteLine(e);
                Assert.IsNotNull(e);
                //DAVtq vtq = e.Vtq;
                //ItemChangedEventArgsList.Add(e)
                //AssertUtilities.WriteLine($"vtq: {vtq}");

                if (e.Succeeded)
                    successCount++;
                else
                    failureCount++;
            }

            client.ItemChanged += ItemChanged;
            try
            {
                AssertUtilities.WriteLine("Subscribing...");

                // ReSharper disable AssignNullToNotNullAttribute
                int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, new DAGroupParameters(100), null);
                // ReSharper restore AssignNullToNotNullAttribute

                AssertUtilities.WriteLine($"handle: {handle}");
                AssertUtilities.WaitingPhase(10 * 1000);

                AssertUtilities.WriteLine("Unsubscribing...");
                if (unsubscribeAllItems)
                    client.UnsubscribeAllItems();
                else
                    client.UnsubscribeItem(handle);
            }
            finally
            {
                client.ItemChanged -= ItemChanged;
            }

            AssertUtilities.WriteLine($"successCount: {successCount}");
            AssertUtilities.WriteLine($"failureCount: {failureCount}");

            if (client.SupportsTechnology(OpcTechnologies.Com))
            {
                Assert.IsTrue(successCount >= 1);
                Assert.AreEqual(0, failureCount);
            }
            else
            {
                Assert.AreEqual(0, successCount);
                Assert.AreEqual(1, failureCount);
            }
        }

        public void SubscribeItemTest_Repeated(
            [NotNull] IEasyDAClient client,
            [CanBeNull] ServerDescriptor serverDescriptor,
            [CanBeNull] DAItemDescriptor itemDescriptor,
            bool unsubscribeAllItems,
            int iterationCount,
            int delayBetweenIterations)
        {
            for (int i = 0; i < iterationCount; i++)
            {
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Iteration #{0}", i));
                if (i != 0)
                    AssertUtilities.WaitingPhase(delayBetweenIterations);
                SubscribeItemTest(client, serverDescriptor, itemDescriptor, unsubscribeAllItems);
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest()
        {
            SubscribeItemTest(
                CreateClient(),
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s),
                unsubscribeAllItems: false);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest_Combined_WithUsing()
        {
            using (var client = (EasyDAClient)CreateClient())
            {
                SubscribeItemTest_Repeated(
                    client,
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s),
                    unsubscribeAllItems: false,
                    iterationCount: 3,
                    delayBetweenIterations: 0);

            }
            Thread.Sleep(2 * 1000);
            using (var client = new EasyDAClient())
            {
                SubscribeItemTest_Repeated(
                    client,
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s),
                    unsubscribeAllItems: false,
                    iterationCount: 3,
                    delayBetweenIterations: 0);
            }
        }

        // Although this test does not contain Assert-s, it tests whether the back-end processing does not collide with
        // Dispose (in which case an exception will be thrown but not handled).
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest_ImmediateDispose()
        {
            using (var client = (EasyDAClient)CreateClient())
            {
                var serverDescriptor = new ServerDescriptor("", OpcLabs_KitServer.ProgId);
                var itemDescriptor = new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s);

                void ItemChanged(object sender, EasyDAItemChangedEventArgs e) => AssertUtilities.WriteLine(e);

                client.ItemChanged += ItemChanged;
                try
                {
                    AssertUtilities.WriteLine("Subscribing...");
                    int handle = client.SubscribeItem(serverDescriptor, itemDescriptor, new DAGroupParameters(100), null);
                    AssertUtilities.WriteLine($"handle: {handle}");
                    AssertUtilities.WaitingPhase(5 * 1000);
                    client.UnsubscribeItem(handle);
                }
                finally
                {
                    client.ItemChanged -= ItemChanged;
                }
            }

            AssertUtilities.WaitingPhase(2*1000);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest_Repeated_3_Immediate()
        {
            SubscribeItemTest_Repeated(
                CreateClient(),
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s),
                unsubscribeAllItems: false,
                iterationCount: 3,
                delayBetweenIterations: 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest_Repeated_3_Delayed()
        {
            SubscribeItemTest_Repeated(
                CreateClient(),
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s),
                unsubscribeAllItems: false,
                iterationCount: 3,
                // ReSharper disable RedundantArgumentName
                delayBetweenIterations: 2 * 1000);
                // ReSharper restore RedundantArgumentName
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest_UnsubscribeAll_3_Immediate()
        {
            SubscribeItemTest_Repeated(
                CreateClient(),
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s),
                unsubscribeAllItems: true,
                iterationCount: 3,
                delayBetweenIterations: 0);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTest_UnsubscribeAll_3_Delayed()
        {
            SubscribeItemTest_Repeated(
                CreateClient(),
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s),
                unsubscribeAllItems: true,
                iterationCount: 3,
                // ReSharper disable RedundantArgumentName
                delayBetweenIterations: 2 * 1000);
                // ReSharper restore RedundantArgumentName
        }

        static private void SubscribeItemTestItemChanged(object sender, EasyDAItemChangedEventArgs e)
        {
            Assert.IsNotNull(e);
            DAVtq vtq = e.Vtq;
            Assert.IsNotNull(vtq);
            //ItemChangedEventArgsList.Add(e)
            AssertUtilities.WriteLine($"value: {vtq.Value}");
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadItemTimestampTest()
        {
            // IMPROVE: Make sure the timezone is not GMT, for the test to be meaningful
            const double millisecondsTolerance = 5000.0;

            var target = CreateClient();
            const int value = 0;
            try
            {
                target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_I4, value);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            DateTime utcNow = DateTime.UtcNow;
            target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_I4, value + 1);
            // change the value

            DAVtq vtq = target.ReadItem("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_I4);
            AssertUtilities.WriteLine(vtq);

            Assert.AreEqual(DateTimeKind.Utc, vtq.Timestamp.Kind);
            TimeSpan span = vtq.Timestamp - utcNow;
            Assert.IsTrue(span.Duration().TotalMilliseconds <= millisecondsTolerance);

            target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_I4, value);
            // revert
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeItemTimestampTest()
        {
            // IMPROVE: Make sure the timezone is not GMT, for the test to be meaningful
            const double millisecondsTolerance = 5000.0;

            var target = CreateClient();
            target.ItemChanged += SubscribeItemTimestampTestItemChanged;

            const int value = 0;
            try
            {
                target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_I4, value);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            DateTime utcNow = DateTime.UtcNow;
            target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_I4, value + 1);
            // change the value

            var argsList = new List<EasyDAItemChangedEventArgs>();
            int actualValue = target.SubscribeItem("", OpcLabs_KitServer.ProgId,
                                                   OpcLabs_KitServer.ItemIds.Register_I4, 100,
                                                   argsList);

            AssertUtilities.WaitingPhase(10 * 1000);

            target.UnsubscribeItem(actualValue);
            target.ItemChanged -= SubscribeItemTimestampTestItemChanged;

            Assert.AreEqual(1, argsList.Count);

            EasyDAItemChangedEventArgs args = argsList[0];
            Assert.IsNotNull(args);
            DAVtq vtq = args.Vtq;
            Assert.IsNotNull(vtq);

            Assert.AreEqual(DateTimeKind.Utc, vtq.Timestamp.Kind);
            TimeSpan span = vtq.Timestamp - utcNow;
            Assert.IsTrue(span.Duration().TotalMilliseconds <= millisecondsTolerance);

            target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_I4, value);
            // revert
        }

        static private void SubscribeItemTimestampTestItemChanged(object sender, EasyDAItemChangedEventArgs e)
        {
            Assert.IsNotNull(e);
            Assert.IsNotNull(e.Arguments.State);
            // ReSharper disable once PossibleNullReferenceException
            ((List<EasyDAItemChangedEventArgs>)e.Arguments.State).Add(e);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeMultipleItemsTest()
        {
            var target = CreateClient();
            var serverDescriptor = new ServerDescriptor("", OpcLabs_KitServer.ProgId);

            target.ItemChanged += SubscribeItemTestItemChanged;

            int[] actualValue = target.SubscribeMultipleItems(new[]
                                                                  {
                                                                      new DAItemGroupArguments(serverDescriptor,
                                                                                               OpcLabs_KitServer.
                                                                                                   ItemIds.
                                                                                                   Incrementing_1s, 100,
                                                                                               null),
                                                                      new DAItemGroupArguments(serverDescriptor,
                                                                                               OpcLabs_KitServer.
                                                                                                   ItemIds.Random, 100,
                                                                                               null)
                                                                  });





            AssertUtilities.WaitingPhase(10 * 1000);

            target.UnsubscribeMultipleItems(actualValue);
            target.ItemChanged -= SubscribeItemTestItemChanged;
        }
        // ReSharper restore InconsistentNaming

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void UnsubscribeAllItemsTest_ItemChangedAbsence()
        {
            UnsubscribeAllItemsTest_ItemChangedAbsence(
                CreateClient,
                // ReSharper disable once PossibleNullReferenceException
                client => client.UnsubscribeAllItems());
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        static private void UnsubscribeAllItemsTest_ItemChangedAbsence(
            [NotNull] Func<IEasyDAClient> createClientFunc,
            [NotNull] Action<IEasyDAClient> unsubscribeAllItemsAction)
        {
            // 
            TestSection.Arrange();

            IEasyDAClient client = createClientFunc();
            Debug.Assert(client != null);
            var serverDescriptor = new ServerDescriptor("", OpcLabs_KitServer.ProgId);
            var itemDescriptor = new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Incrementing_1s);

            var itemChangedEventArgsListSubscribed = new List<EasyDAItemChangedEventArgs>();
            var itemChangedEventArgsListUnsubscribed = new List<EasyDAItemChangedEventArgs>();
            List<EasyDAItemChangedEventArgs> itemChangedEventArgsList = null;

            void UnsubscribeAllItemsTestItemChanged(object sender, EasyDAItemChangedEventArgs e)
            {
                AssertUtilities.WriteLine(e);

                // ReSharper disable once AccessToModifiedClosure
                List<EasyDAItemChangedEventArgs> list = itemChangedEventArgsList;
                Debug.Assert(list != null);
                list.Add(e);

                // IMPROVE: the assertions below do not work as they execute on a separate thread
                Assert.IsNotNull(e);
                DAVtq vtq = e.Vtq;
                Assert.IsNotNull(vtq);
            }

            //
            TestSection.Act();

            AssertUtilities.Phase("Hooking event handler");
            client.ItemChanged += UnsubscribeAllItemsTestItemChanged;

            AssertUtilities.Phase("Subscribing item");
            itemChangedEventArgsList = itemChangedEventArgsListSubscribed;
            client.SubscribeItem(serverDescriptor, itemDescriptor, new DAGroupParameters(100), null);
            AssertUtilities.WaitingPhase(10 * 1000);
            AssertUtilities.Phase("Unsubscribing all items");
            unsubscribeAllItemsAction(client);
            AssertUtilities.WaitingPhase(5 * 1000);
            itemChangedEventArgsList = itemChangedEventArgsListUnsubscribed;
            AssertUtilities.WaitingPhase(10 * 1000);

            AssertUtilities.Phase("Unhooking event handler");
            client.ItemChanged -= UnsubscribeAllItemsTestItemChanged;

            //
            TestSection.Assert();
            int countItemChangedSubscribed = itemChangedEventArgsListSubscribed.Count;
            int countItemChangedUnsubscribed = itemChangedEventArgsListUnsubscribed.Count;
            AssertUtilities.WriteLine($"Item changed count while subscribed: {countItemChangedSubscribed}");
            AssertUtilities.WriteLine($"Item changed count while unsubscribed: {countItemChangedUnsubscribed}");
            if (client.SupportsTechnology(OpcTechnologies.Com))
                Assert.IsTrue(countItemChangedSubscribed >= 2);
            else
                Assert.AreEqual(1, countItemChangedSubscribed);
            Assert.AreEqual(0, countItemChangedUnsubscribed);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteAndReadItemTest_ReadParameters()
        {
            WriteAndReadItemTest(
                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Register_R4),
                new DAReadParameters(DADataSource.Cache, 112233),
                0.0F);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemValueTest_Register_ArrayOfR4()
        {
            var target = CreateClient();
            try
            {
                target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_ArrayOfR4,
                    new[] { 1.1f, 22.2f, 333.3f, 4444.4f, 55555.5f });
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
        }

        // for http://www.opclabs.com/forum/reading-writing-subscriptions-property-access/1407-problem-using-writemultipleitems#2075
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemTest_OnlineForums_1407a()
        {
            const int preWriteValue = 211;
            const int writeValue = 212;
            var writeTimestamp = DateTime.SpecifyKind(new DateTime(2014, 1, 1), DateTimeKind.Utc);
            DAQuality writeQuality = DAQualities.GoodNonspecific;

            // Arrange

            var target = CreateClient();

            try
            {
                target.WriteItemValue(
                    "", OpcLabs_KitServer.ProgId,
                    "Boilers.Boiler #1.Simulation.UpdateRate",
                    preWriteValue);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            // Act

            target.WriteItem(
                "", OpcLabs_KitServer.ProgId, 
                "Boilers.Boiler #1.Simulation.UpdateRate", 
                writeValue, writeTimestamp, writeQuality);

            // Assert

            DAVtq vtq = target.ReadItem(
                "", OpcLabs_KitServer.ProgId,
                "Boilers.Boiler #1.Simulation.UpdateRate");
            Assert.IsNotNull(vtq);
            Assert.AreEqual((Int64)writeValue, vtq.Value);
            Assert.AreEqual(DateTimeKind.Utc, vtq.Timestamp.Kind);
            Assert.AreEqual(writeTimestamp, vtq.Timestamp);
            Assert.AreEqual(writeQuality, vtq.Quality);
        }

        // for http://www.opclabs.com/forum/reading-writing-subscriptions-property-access/1407-problem-using-writemultipleitems#2075
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemTest_OnlineForums_1407b()
        {
            const int preWriteValue = 221;
            var preWriteTimestamp = DateTime.SpecifyKind(new DateTime(2014, 1, 1), DateTimeKind.Utc);
            DAQuality preWriteQuality = DAQualities.GoodNonspecific;
            const int writeValue = 222;

            // Arrange

            var target = CreateClient();

            try
            {
                target.WriteItem(
                    "", OpcLabs_KitServer.ProgId,
                    "Boilers.Boiler #1.Simulation.UpdateRate",
                    preWriteValue, preWriteTimestamp, preWriteQuality);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            // Act

            target.WriteItemValue(
                "", OpcLabs_KitServer.ProgId,
                "Boilers.Boiler #1.Simulation.UpdateRate",
                writeValue);

            // Assert

            DAVtq vtq = target.ReadItem(
                "", OpcLabs_KitServer.ProgId,
                "Boilers.Boiler #1.Simulation.UpdateRate");
            Assert.IsNotNull(vtq);
            Assert.AreEqual((Int64)writeValue, vtq.Value);
            Assert.AreEqual(DateTimeKind.Utc, vtq.Timestamp.Kind);
            Assert.IsTrue(vtq.Timestamp <= DateTime.UtcNow);
            Assert.AreEqual((DAQuality)DAQualities.GoodNonspecific, vtq.Quality);
        }

        // for http://www.opclabs.com/forum/reading-writing-subscriptions-property-access/1407-problem-using-writemultipleitems#2075
        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteItemTest_OnlineForums_1407c()
        {
            const int preWriteValue = 231;
            const int writeValue = 232;
            var writeTimestamp = DateTime.SpecifyKind(new DateTime(2014, 1, 1), DateTimeKind.Utc);
            DAQuality writeQuality = DAQualities.GoodNonspecific;

            // Arrange

            var target = CreateClient();

            try
            {
                target.WriteItemValue(
                    "", OpcLabs_KitServer.ProgId,
                    OpcLabs_KitServer.ItemIds.Register_I4,
                    preWriteValue);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }

            // Act

            target.WriteItem(
                "", OpcLabs_KitServer.ProgId,
                OpcLabs_KitServer.ItemIds.Register_I4,
                writeValue, writeTimestamp, writeQuality);

            // Assert

            DAVtq vtq = target.ReadItem(
                "", OpcLabs_KitServer.ProgId,
                OpcLabs_KitServer.ItemIds.Register_I4);
            Assert.IsNotNull(vtq);
            Assert.AreEqual(writeValue, vtq.Value);
            Assert.AreEqual(DateTimeKind.Utc, vtq.Timestamp.Kind);
            Assert.IsTrue(vtq.Timestamp <= DateTime.UtcNow);
            Assert.AreEqual((DAQuality)DAQualities.GoodNonspecific, vtq.Quality);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        // ReSharper disable InconsistentNaming
        public void WriteItemValueTest_Register_BOOL()
        // ReSharper restore InconsistentNaming
        {
            var target = CreateClient();
            try
            {
                target.WriteItemValue("", OpcLabs_KitServer.ProgId, OpcLabs_KitServer.ItemIds.Register_BOOL, false);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
        }

        public void ReadMultipleItemsTest(int numberOfItems)
        {
            var target = CreateClient();

            var argumentArray = new DAReadItemArguments[numberOfItems];
            for (int i = 0; i < numberOfItems; i++)
            {
                int copy = (i / 100) + 1;
                int phase = i % 100;
                string itemId = String.Format(CultureInfo.InvariantCulture, "Simulation.Incrementing.Copy_{0}.Phase_{1}", copy, phase);
                argumentArray[i] = new DAReadItemArguments("", "OPCLabs.KitServer.2", itemId, 1000);
            }

            DAVtqResult[] actualResultArray = target.ReadMultipleItems(argumentArray);

            Assert.IsNotNull(actualResultArray);
            Assert.AreEqual(1, actualResultArray.Rank);
            Assert.AreEqual(numberOfItems, actualResultArray.Length);

            // TODO:
        }

        [TestCategory("Live")]
        [TestCategory("Load")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemsTest_100_Items()
        {
            ReadMultipleItemsTest(100);
        }

        [TestCategory("Live")]
        [TestCategory("Load")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemsTest_1000_Items()
        {
            ReadMultipleItemsTest(1000);
        }

        [Ignore]    // IMPROVE: put back
        [TestCategory("Live")]
        [TestCategory("Load")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemsTest_5000_Items()
        {
            ReadMultipleItemsTest(5000);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemsTest_Basic()
        {
            var target = CreateClient();

            DAVtqResult[] actualResultArray =
                target.ReadMultipleItems(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new[]
                        {
                            new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Constant_R4),
                            new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Constant_R8)
                        }
                    );

            Assert.IsNotNull(actualResultArray);
            Assert.AreEqual(1, actualResultArray.Rank);
            Assert.AreEqual(2, actualResultArray.Length); //TODO:
            CollectionAssert.AllItemsAreNotNull(actualResultArray);
            CollectionAssert.AllItemsAreInstancesOfType(actualResultArray, typeof(DAVtqResult));

            if (!target.SupportsTechnology(OpcTechnologies.Com))
            {
                foreach (ValueResult valueResult in actualResultArray)
                    // ReSharper disable once PossibleNullReferenceException
                    Assert.IsInstanceOfType(valueResult.Exception, typeof(NotSupportedException));
                return;
            }

            DAVtqResult result0 = actualResultArray[0];
            Assert.IsNotNull(result0);
            Exception exception0 = result0.Exception;
            AssertUtilities.WriteLine(exception0);
            Assert.IsNull(exception0);
            ReadItemAsserts(result0.Vtq, 0.0F);

            DAVtqResult result1 = actualResultArray[1];
            Assert.IsNotNull(result1);
            Assert.IsNull(result1.Exception);
            ReadItemAsserts(result1.Vtq, 0.0D);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void ReadMultipleItemValuesTest_Basic()
        {
            var target = CreateClient();

            ValueResult[] actualResultArray =
                target.ReadMultipleItemValues(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new[]
                        {
                            new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Constant_R4),
                            new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Constant_R8)
                        }
                    );

            Assert.IsNotNull(actualResultArray);
            Assert.AreEqual(1, actualResultArray.Rank);
            Assert.AreEqual(2, actualResultArray.Length); //TODO:
            CollectionAssert.AllItemsAreNotNull(actualResultArray);
            CollectionAssert.AllItemsAreInstancesOfType(actualResultArray, typeof(ValueResult));

            if (!target.SupportsTechnology(OpcTechnologies.Com))
            {
                foreach (ValueResult valueResult in actualResultArray)
                    // ReSharper disable once PossibleNullReferenceException
                    Assert.IsInstanceOfType(valueResult.Exception, typeof(NotSupportedException));
                return;
            }

            ValueResult result0 = actualResultArray[0];
            Assert.IsNotNull(result0);
            Assert.IsNull(result0.Exception);
            Assert.AreEqual(0.0F, result0.Value);

            ValueResult result1 = actualResultArray[1];
            Assert.IsNotNull(result1);
            Assert.IsNull(result1.Exception);
            Assert.AreEqual(0.0D, result1.Value);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteMultipleItemsTest_Basic()
        {
            var target = CreateClient();

            OperationResult[] actualResultArray =
                target.WriteMultipleItems(
                    new[]
                        {
                            new DAItemVtqArguments(
                                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Register_R4),
                                new DAVtq(0.0F), null),
                            new DAItemVtqArguments(
                                new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                                new DAItemDescriptor(OpcLabs_KitServer.ItemIds.Register_R8),
                                new DAVtq(0.0D), null)
                        }
                    );

            Assert.IsNotNull(actualResultArray);
            Assert.AreEqual(1, actualResultArray.Rank);
            Assert.AreEqual(2, actualResultArray.Length); //TODO:
            CollectionAssert.AllItemsAreNotNull(actualResultArray);
            CollectionAssert.AllItemsAreInstancesOfType(actualResultArray, typeof(OperationResult));

            foreach (OperationResult operationResult in actualResultArray)
            {
                Assert.IsNotNull(operationResult);
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(operationResult.Exception, typeof(NotSupportedException));
                    continue;
                }
                Assert.IsNull(operationResult.Exception);
            }
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteMultipleItemsTest_ReadBack()
        {
            var target = CreateClient();

            const int count = 8;
            var argumentsArray = new DAItemValueArguments[count];
            var random = new Random();
            for (int i = 0; i < count; i++)
            {
                int value = random.Next();
                AssertUtilities.WriteLine(value);
                argumentsArray[i] = new DAItemValueArguments(
                    OpcLabs_KitServer.ProgId,
                    String.Format(CultureInfo.InvariantCulture, "Simulation.RegisterSet_{0}.I4", i + 1),
                    value,
                    null);
            }

            OperationResult[] actualResultArray = target.WriteMultipleItemValues(argumentsArray);

            Assert.IsNotNull(actualResultArray);
            Assert.AreEqual(1, actualResultArray.Rank);
            Assert.AreEqual(count, actualResultArray.Length);
            CollectionAssert.AllItemsAreNotNull(actualResultArray);
            CollectionAssert.AllItemsAreInstancesOfType(actualResultArray, typeof(OperationResult));

            foreach (OperationResult operationResult in actualResultArray)
            {
                Assert.IsNotNull(operationResult);
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(operationResult.Exception, typeof(NotSupportedException));
                    continue;
                }
                Assert.IsNull(operationResult.Exception);
            }

            if (!target.SupportsTechnology(OpcTechnologies.Com))
                return;

            ValueResult[] readBackResultArray = target.ReadMultipleItemValues(
                argumentsArray.Select(arguments => new DAReadItemArguments(
                    // ReSharper disable PossibleNullReferenceException
                    arguments.ServerDescriptor,
                    // ReSharper restore PossibleNullReferenceException
                    arguments.ItemDescriptor,
                    new DAReadParameters(valueAge: 0))).ToArray());

            Assert.IsNotNull(readBackResultArray);
            Assert.AreEqual(1, readBackResultArray.Rank);
            Assert.AreEqual(count, readBackResultArray.Length);
            CollectionAssert.AllItemsAreNotNull(readBackResultArray);
            CollectionAssert.AllItemsAreInstancesOfType(readBackResultArray, typeof(OperationResult));
            // ReSharper disable PossibleNullReferenceException
            Assert.IsTrue(readBackResultArray.All(result => result.Exception == null));
            // ReSharper restore PossibleNullReferenceException

            for (int i = 0; i < count; i++)
                // ReSharper disable PossibleNullReferenceException
                Assert.AreEqual(argumentsArray[i].Value, readBackResultArray[i].Value,
                    "Value mismatch at index {0}.", i);
            // ReSharper restore PossibleNullReferenceException
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void WriteMultipleItemsTest_ReadBack_Repeated_100()
        {
            WriteMultipleItemsTest_ReadBack_Repeated(100, 0);
        }



        #region Protected Methods

        protected void ReadItemAsserts(DAVtq actualVtq, object expectedValue)
        {
            Assert.IsNotNull(actualVtq);

            Assert.IsTrue(actualVtq.Quality.IsGood);
            Assert.AreEqual(DAStatusChoice.GoodNonspecific, actualVtq.Quality.StatusBitField);
            Assert.AreEqual(DALimitChoice.LimitOk, actualVtq.Quality.LimitBitField);

            Assert.AreEqual(DateTimeKind.Utc, actualVtq.Timestamp.Kind);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MinValue);
            Assert.AreNotEqual(actualVtq.Timestamp, DateTime.MaxValue);
            Assert.IsTrue(actualVtq.Timestamp >= DateTime.FromFileTimeUtc(0));
            Assert.IsTrue(actualVtq.Timestamp <= DateTime.Now);

            Assert.AreEqual(expectedValue, actualVtq.Value);
        }

        protected bool ReadItemTest([NotNull] string machineName, [NotNull] string serverClass, [NotNull] string itemId,
            [NotNull] object expectedValue)
        {
            var target = CreateClient();
            DAVtq actualVtq;
            try
            {
                actualVtq = target.ReadItem(machineName, serverClass, itemId);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return false;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualVtq);
            ReadItemAsserts(actualVtq, expectedValue);
            return true;
        }

        protected void ReadItemTest([NotNull] string machineName, [NotNull] string serverClass, [NotNull] string itemId,
            short dataType, [CanBeNull] object expectedValue)
        {
            var target = CreateClient();
            DAVtq actualVtq;
            try
            {
                actualVtq = target.ReadItem(machineName, serverClass, itemId, dataType);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualVtq);
            ReadItemAsserts(actualVtq, expectedValue);
        }

        protected void ReadItemTest([NotNull] ServerDescriptor serverDescriptor, [NotNull] DAItemDescriptor itemDescriptor,
                                    [CanBeNull] object expectedValue)
        {
            var target = CreateClient();
            DAVtq actualVtq;
            try
            {
                actualVtq = target.ReadItem(serverDescriptor, itemDescriptor);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            AssertUtilities.WriteLine(actualVtq);
            ReadItemAsserts(actualVtq, expectedValue);
        }

        protected void WriteAndReadItemTest([NotNull] ServerDescriptor serverDescriptor, [NotNull] DAItemDescriptor itemDescriptor,
                                    [NotNull] DAReadParameters readParameters, object expectedValue)
        {
            var target = CreateClient();
            try
            {
                target.WriteItemValue(serverDescriptor, itemDescriptor, expectedValue);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                throw;
            }
            DAVtq actualVtq = target.ReadItem(serverDescriptor, itemDescriptor, readParameters);
            AssertUtilities.WriteLine(actualVtq);
            ReadItemAsserts(actualVtq, expectedValue);
        }

        #endregion



        // ReSharper disable InconsistentNaming
        private void WriteMultipleItemsTest_ReadBack_Repeated(int iterationCount, int delayBetweenIterations)
        // ReSharper restore InconsistentNaming
        {
            for (int i = 0; i < iterationCount; i++)
            {
                AssertUtilities.WriteLine(String.Format(CultureInfo.InvariantCulture, "Iteration {0}", i));
                if (i != 0)
                    AssertUtilities.WaitingPhase(delayBetweenIterations);
                WriteMultipleItemsTest_ReadBack();
            }
        }
    }
}
