﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Extensions;
using OpcLabs.EasyOpc.OperationModel;
using OpcLabs.EasyOpcLiveTest.DataAccess.Environment;

namespace OpcLabs.EasyOpcLiveTest.DataAccess
{
    // ReSharper disable once InconsistentNaming
    public class IEasyDAClientTest_Com_BrowseAccessPaths : IEasyDAClientTestBase
    {
        public IEasyDAClientTest_Com_BrowseAccessPaths([NotNull] Func<IEasyDAClient> createClientFunc)
            : base(createClientFunc)
        {
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseAccessPathsTest()
        {
            var target = CreateClient();

            string[] accessPathsArray;
            try
            {
                accessPathsArray = target.BrowseAccessPaths("", OpcLabs_KitServer.ProgId,
                    OpcLabs_KitServer.ItemIds.Constant_I4);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                if (opcException.InnerException is NotImplementedException)
                    return;
                throw;
            }

            Assert.IsNotNull(accessPathsArray);
            Assert.AreEqual(2, accessPathsArray.Length);

            //TODO:

        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseAccessPathsTest_ServerDescriptor()
        {
            var target = CreateClient();

            string[] accessPathsArray;
            try
            {
                accessPathsArray = target.BrowseAccessPaths(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId), OpcLabs_KitServer.ItemIds.Constant_I4);
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                if (opcException.InnerException is NotImplementedException)
                    return;
                throw;
            }

            Assert.IsNotNull(accessPathsArray);
            Assert.AreEqual(2, accessPathsArray.Length);

            //TODO:

        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void BrowseAccessPathsTest_BrowsePath()
        {
            var target = CreateClient();

            string[] accessPathsArray;
            try
            {
                accessPathsArray = target.BrowseAccessPaths(
                    new ServerDescriptor("", OpcLabs_KitServer.ProgId),
                    new DANodeDescriptor(null, OpcLabs_KitServer.BrowsePaths.Constant_I4));
            }
            catch (OpcException opcException)
            {
                if (!target.SupportsTechnology(OpcTechnologies.Com))
                {
                    Assert.IsInstanceOfType(opcException.InnerException, typeof(NotSupportedException));
                    return;
                }
                if (opcException.InnerException is NotImplementedException)
                    return;
                throw;
            }

            Assert.IsNotNull(accessPathsArray);
            Assert.AreEqual(2, accessPathsArray.Length);

            //TODO:

        }
    }
}
