﻿using System.Collections.Generic;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA.DataTypeModel
{
    [TestClass]
    public class DataTypeDefinitionTest
    {
        #region Static Private Properties

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static private IEnumerable<UANodeDescriptor> NodeDescriptors_UnifiedAutomation_UaCppServer =>
            new UANodeDescriptor[]
            {
                "ns=2;s=Demo.Static.Scalar.CarExtras",
                "ns=2;s=Demo.Static.Scalar.OptionSet",
                //"ns=2;s=Demo.Static.Scalar.Structure",
                "ns=2;s=Demo.Static.Scalar.Structures.AnalogMeasurement",
                "ns=2;s=Demo.Static.Scalar.Structures.StatusSymbols",
                "ns=2;s=Person1",
                "ns=2;s=Person2",
                "ns=2;s=Person3",
            };
        
        #endregion



        #region Test Methods

        [TestMethod]
        public void ReadTest_UnifiedAutomation_UaCppServer()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.UnifiedAutomation_UaCppServer.EndpointDescriptors.OpcTcp;
            endpointDescriptor.Host = "localhost";  // TODO: remove
            AssertUtilities.WriteLine($"endpointDescriptor: {endpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var client = new EasyUAClient();

            foreach (UANodeDescriptor nodeDescriptor in NodeDescriptors_UnifiedAutomation_UaCppServer)
            {
                TestSection.Act();

                AssertUtilities.WriteLine($"nodeDescriptor: {nodeDescriptor}");
                object value = client.ReadValue(endpointDescriptor, nodeDescriptor);

                TestSection.Assert();

                AssertUtilities.WriteLine($"value: {value}");
                Assert.IsNotNull(value);

                // IMPROVE: more Checks
            }
        }

        [TestMethod]
        public void ReadWriteTest_UnifiedAutomation_UaCppServer()
        {
            TestSection.Arrange();

            UAEndpointDescriptor endpointDescriptor = UALiveTestEnvironment.UnifiedAutomation_UaCppServer.EndpointDescriptors.OpcTcp;
            endpointDescriptor.Host = "localhost";  // TODO: remove
            AssertUtilities.WriteLine($"endpointDescriptor: {endpointDescriptor}");
            TestEnvironment.PrecheckEndpoint(endpointDescriptor);

            var client = new EasyUAClient();

            foreach (UANodeDescriptor nodeDescriptor in NodeDescriptors_UnifiedAutomation_UaCppServer)
            {
                TestSection.Act();

                AssertUtilities.WriteLine($"nodeDescriptor: {nodeDescriptor}");
                object value = client.ReadValue(endpointDescriptor, nodeDescriptor);

                TestSection.Assert();

                AssertUtilities.WriteLine($"value: {value}");
                Assert.IsNotNull(value);

                client.WriteValue(endpointDescriptor, nodeDescriptor, value);
            }
        }

        #endregion



        #region Private Properties

        [NotNull]
        private UALiveTestEnvironment TestEnvironment => UALiveTestEnvironment.Current;

        #endregion
    }
}
