﻿
using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.BaseLib.Collections.Generic;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.ComponentModel;
using OpcLabs.BaseLib.Net.Http;
using OpcLabs.BaseLib.Net.Sockets;
using OpcLabs.BaseLib.Threading;
using OpcLabs.BaseLib.Utilities;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.Discovery;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpcLiveTest.UA.Environment;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA
{
    [TestClass]
    public class UALiveTestEnvironment
    {
        #region Static Constructor

        static UALiveTestEnvironment()
        {
            LicensingManagement.Instance.RegisterManagedResource("QuickOPC", "Multipurpose",
                Assembly.GetExecutingAssembly(),
                "OpcLabs.EasyOpcLiveTest.Key-Evaluation-WebForm-1999003411-20670827.bin");
            //var client = new EasyUAClient();
            //long serialNumber = (uint)client.LicenseInfo["Multipurpose.SerialNumber"];
        }

        #endregion



        #region Public Constructors

        public UALiveTestEnvironment()
        {
            EasyUAClient.SharedParameters = EasyUASharedParameters.TestingInsecure;

            UAServer Classifier(UAEndpointDescriptor endpointDescriptor)
            {
                return AllServers.First(server =>
                    // ReSharper disable once PossibleNullReferenceException
                    server.AllEndpointDescriptors.Any(descriptor => descriptor == endpointDescriptor));
            }

            _monitorLockPool = new MonitorLockPool<UAEndpointDescriptor>(
                new ClassifyingEqualityComparer<UAEndpointDescriptor, UAServer>(
                    Classifier, 
                    ReferenceEqualityComparer<UAServer>.Default));
        }

        #endregion



        #region Static Public Properties

        #endregion



        #region Static Public Methods

        static public bool IsSupportedOptimistic([NotNull] UAEndpointDescriptor endpointDescriptor) // move elsewhere!
        {
#if !(NETFRAMEWORK)    // TODO:
            string scheme = endpointDescriptor.Scheme;
            if (scheme == null)
                return true;
            
            if (UriUtilities.SchemesEqual(scheme, UASchemes.Http))
                return false;
            if (UriUtilities.SchemesEqual(scheme, UASchemes.Https))
            {
                if (EasyUAClient.SharedParameters.EngineParameters.UseCustomHttpsCertificateValidation &&
                    !HttpClientHandlerUtilities.IsCustomCertificateValidationSupported)
                    return false;
            }
#endif
            return true;
        }
        
        #endregion



        #region Static Public Fields

        [NotNull]
        static public readonly UALiveTestEnvironment Current = new UALiveTestEnvironment();

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public OpcFoundation_GlobalDiscoveryServer OpcFoundation_GlobalDiscoveryServer = new OpcFoundation_GlobalDiscoveryServer(UALiveTestDefinitions.HostNames.OpcUATest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<UAEndpointDescriptor[]> OpcFoundation_GlobalDiscoveryServer_AllEndpointDescriptorDataRows
        {
            get
            {
                return OpcFoundation_GlobalDiscoveryServer.AllEndpointDescriptors
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public OpcFoundation_UAAlarmConditionServer OpcFoundation_UAAlarmConditionServer = new OpcFoundation_UAAlarmConditionServer(UALiveTestDefinitions.HostNames.OpcUATest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<UAEndpointDescriptor[]> OpcFoundation_UAAlarmConditionServer_AllEndpointDescriptorDataRows
        {
            get
            {
                return OpcFoundation_UAAlarmConditionServer.AllEndpointDescriptors
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public OpcFoundation_UAReferenceServer OpcFoundation_UAReferenceServer = new OpcFoundation_UAReferenceServer(UALiveTestDefinitions.HostNames.OpcUATest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<UAEndpointDescriptor[]> OpcFoundation_UAReferenceServer_AllEndpointDescriptorDataRows
        {
            get
            {
                return OpcFoundation_UAReferenceServer.AllEndpointDescriptors
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public OpcFoundation_UASampleServer OpcFoundation_UASampleServer = new OpcFoundation_UASampleServer(UALiveTestDefinitions.HostNames.OpcUATest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<UAEndpointDescriptor[]> OpcFoundation_UASampleServer_AllEndpointDescriptorDataRows
        {
            get
            {
                return OpcFoundation_UASampleServer.AllEndpointDescriptors
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<UAEndpointDescriptor[]> OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows
        {
            get
            {
                return OpcFoundation_UASampleServer.AllEndpointDescriptors
                    .Where(IsSupportedOptimistic)
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        // ReSharper disable once IdentifierTypo
        static public Prosys_UASimulationServer Prosys_UASimulationServer = new Prosys_UASimulationServer(UALiveTestDefinitions.HostNames.OpcUATest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        // ReSharper disable once IdentifierTypo
        static public IEnumerable<UAEndpointDescriptor[]> Prosys_UASimulationServer_AllEndpointDescriptorDataRows
        {
            get
            {
                return Prosys_UASimulationServer.AllEndpointDescriptors
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        // ReSharper disable once IdentifierTypo
        static public Softing_OpcUaNetDemoServer Softing_OpcUaNetDemoServer = new Softing_OpcUaNetDemoServer(UALiveTestDefinitions.HostNames.OpcUATest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        // ReSharper disable once IdentifierTypo
        static public IEnumerable<UAEndpointDescriptor[]> Softing_OpcUaNetDemoServer_AllEndpointDescriptorDataRows
        {
            get
            {
                return Softing_OpcUaNetDemoServer.AllEndpointDescriptors
                    .Select(descriptor => new[] { descriptor });
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<UAEndpointDescriptor[]> OpcFoundation_UASampleServer_SynthesizableEndpointDescriptorDataRows
        {
            get
            {
                return OpcFoundation_UASampleServer_OptimisticSupportedEndpointDescriptorDataRows
                    // ReSharper disable PossibleNullReferenceException
                    .Select(dataRow => dataRow[0])
                    // ReSharper disable once AssignNullToNotNullAttribute
                    .Select(descriptor => new UAEndpointDescriptor(descriptor)
                    {
                        UrlString = UriUtilities.HasScheme(descriptor.UrlString, UASchemes.Http)
                            ? descriptor.UrlString + "/None"
                            : descriptor.UrlString,
                        EndpointSelectionPolicy = UAEndpointSelectionPolicy.NoMessageSecurity
                    })
                    .Select(descriptor => new[] {descriptor});
                // ReSharper restore PossibleNullReferenceException
            }
        }

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public UnifiedAutomation_UaCppServer UnifiedAutomation_UaCppServer = new UnifiedAutomation_UaCppServer(UALiveTestDefinitions.HostNames.OpcUATest);

        [NotNull]
        // ReSharper disable once InconsistentNaming
        static public IEnumerable<UAEndpointDescriptor[]> UnifiedAutomation_UaCppServer_AllEndpointDescriptorDataRows
        {
            get
            {
                return UnifiedAutomation_UaCppServer.AllEndpointDescriptors
                    .Select(descriptor => new[] { descriptor });
            }
        }

        #endregion



        #region Public Properties

        [NotNull]
        public IEnumerable<UAServer> AllServers
        {
            get
            {
                yield return OpcFoundation_UAAlarmConditionServer;
                yield return OpcFoundation_UAReferenceServer;
                yield return OpcFoundation_UASampleServer;
                yield return Prosys_UASimulationServer;
                yield return Softing_OpcUaNetDemoServer;
                yield return UnifiedAutomation_UaCppServer;
            }
        }

        #endregion



        #region Public Methods

        public bool CanConnectEndpoint([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            return TestEndpoint(endpointDescriptor) == 0;
        }

        public void PrecheckEndpoint([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            AssertUtilities.Phase($"Pre-checking endpoint \"{endpointDescriptor}\"");

            int error = TestEndpoint(endpointDescriptor);
            if (error != 0)
            {
                string message = $"Endpoint \"{endpointDescriptor}\" pre-check failed: {error}";
                AssertUtilities.WriteLine(message);
                Assert.Inconclusive(message);
            }
        }

        public void PrecheckLds([NotNull] string hostName)
        {
            UAEndpointDescriptor ldsEndpointDescriptor = new UAEndpointDescriptor { Host = hostName, Port = 4840 };
            PrecheckEndpoint(ldsEndpointDescriptor);
        }

        [NotNull]
        public IDisposable RequestLockFor([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            return _monitorLockPool.RequestLockFor(endpointDescriptor);
        }

        public int TestEndpoint([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            int error;
            lock (_testEndpointCache)
            {
                if (!_testEndpointCache.TryGetValue(endpointDescriptor, out error))
                {
                    error = InternalTestEndpoint(endpointDescriptor);
                    _testEndpointCache.AddNew(endpointDescriptor, error);
                }
            }
            return error;
        }

        #endregion



        #region Test Methods

        /// <summary>
        /// Summarizes the availability of OPC UA server endpoints for tests.
        /// </summary>
        // IMPROVE: [DynamicData(nameof(OpcFoundation_GlobalDiscoveryServer_AllEndpointDescriptorDataRows))]
        // IMPROVE: [DynamicData(nameof(OpcFoundation_UAAlarmConditionServer_AllEndpointDescriptorDataRows))]
        // IMPROVE: [DynamicData(nameof(OpcFoundation_UAReferenceServer_AllEndpointDescriptorDataRows))]
        // IMPROVE: [DynamicData(nameof(OpcFoundation_UASampleServer_AllEndpointDescriptorDataRows))]
        // IMPROVE: [DynamicData(nameof(Prosys_UASimulationServer_AllEndpointDescriptorDataRows))]
        // IMPROVE: [DynamicData(nameof(Softing_OpcUaNetDemoServer_AllEndpointDescriptorDataRows))]
        // IMPROVE: [DynamicData(nameof(UnifiedAutomation_UaCppServer_AllEndpointDescriptorDataRows))]
        [TestMethod]
        public void AvailabilityTest_ServerEndpoint([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            PrecheckEndpoint(endpointDescriptor);
        }

        #endregion



        #region Public Fields

        #endregion



        #region Private Methods

        private int InternalTestEndpoint([NotNull] UAEndpointDescriptor endpointDescriptor)
        {
            return TcpClientUtilities.TestCreate(endpointDescriptor, millisecondsTimeout: 1500);
        }

        #endregion



        #region Private Fields

        [NotNull] 
        private readonly MonitorLockPool<UAEndpointDescriptor> _monitorLockPool;

        [NotNull]
        private readonly Dictionary<EndpointDescriptor, int> _testEndpointCache = new Dictionary<EndpointDescriptor, int>();

        #endregion
    }
}
