﻿
using Microsoft.Extensions.DependencyInjection;
using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Browsing;
using OpcLabs.BaseLib.Connectivity.Point;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.EasyOpc.DataAccess.Connectivity;
using OpcLabs.EasyOpc.DataAccess.Connectivity.Internal;

namespace OpcLabs.EasyOpcTest.DataAccess.Connectivity
{
    public partial class DAConnectivityTest
    {
        #region Static Private Methods

        static private void GetServiceTest_Failure([NotNull] Type serviceType)
        {
            var connectivity = new DAConnectivity();
            object service = connectivity.GetService(serviceType);
            Assert.IsNull(service);
        }

        static private void GetServiceTest_Success([NotNull] Type serviceType)
        {
            var connectivity = new DAConnectivity();
            object service = connectivity.GetService(serviceType);
            Assert.IsNotNull(service);
            Assert.IsInstanceOfType(service, serviceType);
        }

        #endregion



        #region Test Methods

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void CreatePointBrowseableTest()
        {
            var connectivity = new DAConnectivity();
            // ReSharper disable once PossibleNullReferenceException
            IBrowseable browseable = connectivity.GetService<IBrowsePointService>().CreatePointBrowseable(new object());
            Assert.IsNotNull(browseable);
            // IMPROVE: test more
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void CreateReaderTest()
        {
            var connectivity = new DAConnectivity();
            // ReSharper disable once PossibleNullReferenceException
            var pointReader = (DAPointReader) connectivity.GetService<IReadPointService>().CreateReader();
            Assert.IsNotNull(pointReader);
            Assert.IsFalse(pointReader.IsFrozen);
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetServiceTest_ICreateParametersService()
        {
            GetServiceTest_Success(typeof(ICreateParametersService));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetServiceTest_ICreatePointService()
        {
            GetServiceTest_Success(typeof(ICreatePointService));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetServiceTest_IBrowsePointService()
        {
            GetServiceTest_Success(typeof(IBrowsePointService));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetServiceTest_IReadPointService()
        {
            GetServiceTest_Success(typeof(IReadPointService));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetServiceTest_ISubscribePointService()
        {
            GetServiceTest_Success(typeof(ISubscribePointService));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetServiceTest_IWritePointService()
        {
            GetServiceTest_Success(typeof(IWritePointService));
        }

        [TestCategory("OpcClassic")]
        [TestCategory("OpcDataAccess")]
        [TestMethod]
        public void GetServiceTest_Object()
        {
            GetServiceTest_Failure(typeof(Object));
        }

        #endregion
    }
}
