﻿
using System;
using System.Diagnostics;
using System.IO;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.DataTypeModel;
using OpcLabs.BaseLib.IO.Extensions;
using OpcLabs.EasyOpc.UA.DataTypeModel.Implementation;

namespace OpcLabs.EasyOpcTest.UA.DataTypeModel.Internal
{
    [TestClass]
    public partial class DeveloperTest : UAOpcBinaryDataTypeSystemTest
    {
        #region Static Private Properties

        [NotNull]
        static private byte[] TypeDictionaryByteArray
        {
            get
            {
                Type type = typeof(UAOpcBinaryDataTypeSystemTest);
                Stream stream = type.Assembly.GetManifestResourceStream(type, "DeveloperTypeDictionary.xml");
                Debug.Assert(stream != null);
                byte[] dictionaryByteArray = stream.ReadToArrayOfBytes();
                return dictionaryByteArray;
            }
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void DecodeErrorTest_BinaryEncodingSwitchFieldNotInStructuredType() =>
            DecodeErrorTest(null, TypeDictionaryByteArray, "DataValue2", new byte[1], 
                new [] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryDecoderFailure, UADataTypeModelErrorCodes.OpcBinaryEncodingSwitchFieldNotInStructuredType});

        [TestMethod]
        public void DecodeErrorTest_BinarySystemDataTypeDescriptionDuplicate() =>
            DecodeErrorTest(null, TypeDictionaryByteArray, "IntegerArray", IntegerArray_0ExtensionObjectBody, UADataTypeModelErrorCodes.OpcBinarySystemDataTypeDescriptionDuplicate);

        [TestMethod]
        public void DecodeErrorTest_BinarySystemNoSatisfyingImportDirective() =>
            DecodeErrorTest(null, TypeDictionaryByteArray, "DataValue1", new byte[1], UADataTypeModelErrorCodes.OpcBinarySystemNoSatisfyingImportDirective);

        [TestMethod]
        public void DecoderErrorTest_BinarySystemTypeNameReferenceCircular() =>
            // ReSharper disable once PossibleNullReferenceException
            DecodeErrorTest(dts => dts.Parameters.AllowCircularTypeNameReferences = false, 
                TypeDictionaryByteArray, "StructuredCircular1", new byte[0], UADataTypeModelErrorCodes.OpcBinarySystemTypeNameReferenceCircular);

        [TestMethod]
        public void DecoderErrorTest_BinarySystemTypeNameReferenceCircular_2() =>
            // ReSharper disable once PossibleNullReferenceException
            DecodeErrorTest(dts => dts.Parameters.AllowCircularTypeNameReferences = false,
                TypeDictionaryByteArray, "StructuredCircular2", new byte[0], UADataTypeModelErrorCodes.OpcBinarySystemTypeNameReferenceCircular);

        [TestMethod]
        public void DecoderErrorTest_BinarySystemTypeNameReferenceCircular_3() =>
            // ReSharper disable once PossibleNullReferenceException
            DecodeErrorTest(dts => dts.Parameters.AllowCircularTypeNameReferences = false,
                TypeDictionaryByteArray, "SequenceCircular1", new byte[0], UADataTypeModelErrorCodes.OpcBinarySystemTypeNameReferenceCircular);

        [TestMethod]
        public void EncodeErrorTest_BinaryEncoderSequenceDataActualLengthMismatch() =>
            EncodeErrorTest(null, 
                TypeDictionaryByteArray,
                "IntegerArrayWithFixedLength",
                IntegerArrayWithFixedLength_2Data,
                new [] { UADataTypeModelErrorCodes.OpcBinarySystemDictionaryEncoderFailure, UADataTypeModelErrorCodes.OpcBinaryDictionaryEncoderSequenceDataActualLengthMismatch});

        // IMPROVE: move elsewhere
        [ExpectedException(typeof(ArgumentNullException))]
        [TestMethod]
        public void StructuredDataTest_ArgumentChecking_GenericDataNull()
        {
            // ReSharper disable once ObjectCreationAsStatement
            new StructuredData { { "aName", null } };
        }

        #endregion
    }
}
