﻿using System;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpcFactory.UA;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.UA
{
    public partial class UAAttributeDataTest
    {
        #region Static Public Methods

        [NotNull]
        static public UAAttributeData CreateSerializationTarget0()
        {
            return new UAAttributeData();
        }

        [NotNull]
        static public UAAttributeData CreateSerializationTarget1()
        {
            return new UAAttributeData("value1", UAStatusCode.FromUASeverity(UASeverity.UncertainOrWarning),
                                       DateTime.MinValue, DateTime.MaxValue);
        }

        #endregion



        #region Test Methods

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BinarySerializationTest0()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("BinarySerialization")]
        [TestCategory("OpcUA")]
        [TestMethod]
        public void BinarySerializationTest1()
        {
            Tests.BinarySerializationTest(CreateSerializationTarget1());
        }

        [TestCategory("OpcUA")]
        [TestMethod]
        public void DisplayValueTest0()
        {
            var target = new UAAttributeData();
            Assert.AreEqual("", target.DisplayValue());
        }

        [TestCategory("OpcUA")]
        [TestMethod]
        public void DisplayValueTest1()
        {
            var target = new UAAttributeData(value: 1);
            Assert.AreEqual("1", target.DisplayValue());
        }

        [TestCategory("OpcUA")]
        [TestMethod]
        public void ToStringTest0()
        {
            var target = new UAAttributeData();
            Assert.AreEqual("null {} @0001-01-01T00:00:00.000 @@0001-01-01T00:00:00.000; Good", target.ToString());
        }

        [TestCategory("OpcUA")]
        [TestMethod]
        public void ToStringTest1()
        {
            // ReSharper disable once RedundantArgumentName
            var target = new UAAttributeData(value: null, statusCode:UAStatusCode.FromUASeverity(UASeverity.BadOrFailure));
            Assert.AreEqual("@0001-01-01T00:00:00.000 @@0001-01-01T00:00:00.000; Bad", target.ToString());
        }

        [TestCategory("OpcUA")]
        [TestMethod]
        public void ToStringTest2()
        {
            var target = new UAAttributeData(value: 1);
            Assert.AreEqual("1 {Int32} @0001-01-01T00:00:00.000 @@0001-01-01T00:00:00.000; Good", target.ToString());
        }

        [TestCategory("OpcUA")]
        [TestMethod]
        public void ToStringTest3()
        {
            // ReSharper disable once RedundantArgumentName
            var target = UAAttributeDataInstances.Instance3;
            Assert.AreEqual("222 {Int32} @2014-09-23T14:02:33.555 @@2015-05-14T17:12:34.556; Good", target.ToString());
        }

        [TestCategory("OpcUA")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest0()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget0());
        }

        [TestCategory("OpcUA")]
        [TestCategory("XmlSerialization")]
        [TestMethod]
        public void XmlSerializationTest1()
        {
            Tests.XmlSerializationTest(CreateSerializationTarget1());
        }

        #endregion
    }
}
